/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <math.h>
#include <string.h>

#include "Common.h"
#include "ix86/ix86.h"
#include "Vif.h"
#include "VU0.h"
#include "VUmicro.h"
#include "VifDma.h" 
#ifdef __i386__
#include "i386.h"
#endif

// Extern variables
extern VIFregisters *_vifRegs;
extern int cl;
extern int wl;

// Generic constants
static const unsigned int VIF0intc = 4;
static const unsigned int VIF1intc = 5;
static const unsigned int VIF0dmanum = 0;
static const unsigned int VIF1dmanum = 1;

DMACh *vif0ch;
DMACh *vif1ch;

static int cycles;

typedef void (*UNPACKFUNCTYPE)( u32 *dest, u32 *data );
typedef int  (*UNPACKPARTFUNCTYPE)( u32 *dest, u32 *data, int size );

typedef struct {
	UNPACKFUNCTYPE       funcU;
	UNPACKFUNCTYPE       funcS;
	UNPACKPARTFUNCTYPE   funcUpart;
	UNPACKPARTFUNCTYPE   funcSpart;

	int bsize;
	int dsize;
	int gsize;
} VIFUnpackFuncTable;

#define _UNPACK_TABLE32(name, bsize, dsize, gsize) \
   { UNPACK_##name,         UNPACK_##name, \
     UNPACK_##name##part,   UNPACK_##name##part, \
	 bsize, dsize, gsize },

#define _UNPACK_TABLE(name, bsize, dsize, gsize) \
   { UNPACK_##name##u,      UNPACK_##name##s, \
     UNPACK_##name##upart,  UNPACK_##name##spart, \
	 bsize, dsize, gsize },

// Main table for function unpacking
VIFUnpackFuncTable VIFfuncTable[16] = {
	// 0x0 - S-32
	_UNPACK_TABLE32(S_32, 12, 4, 4)
	// 0x1 - S-16
	_UNPACK_TABLE(S_16, 6, 2, 4)
	// 0x2 - S-8
	_UNPACK_TABLE(S_8, 3, 1, 4)
	// 0x3
   { NULL, NULL, NULL, NULL, 
	 0, 0, 0 },
	// 0x4 - V2-32
	_UNPACK_TABLE32(V2_32, 24, 4, 1)
	// 0x5 - V2-16
	_UNPACK_TABLE(V2_16, 12, 2, 1)
	// 0x6 - V2-8
	_UNPACK_TABLE(V2_8, 6, 1, 1)
	// 0x7
   { NULL, NULL, NULL, NULL,
	 0, 0, 0 },
	// 0x8 - V3-32
	_UNPACK_TABLE32(V3_32, 36, 4, 1)
	// 0x9 - V3-16
	_UNPACK_TABLE(V3_16, 18, 2, 1)
	// 0xA - V3-8
	_UNPACK_TABLE(V3_8, 9, 1, 1)
	// 0xB
   { NULL, NULL, NULL, NULL,
	 0, 0, 0 },
	// 0xC - V4-32
	_UNPACK_TABLE32(V4_32, 48, 4, 1)
	// 0xD - V4-16
	_UNPACK_TABLE(V4_16, 24, 2, 1)
	// 0xE - V4-8
	_UNPACK_TABLE(V4_8, 12, 1, 1)
   // 0xF - V4-5
	_UNPACK_TABLE32(V4_5, 6, 2, 1)
};


void vif0FLUSH();
void vif1FLUSH();

void vifDmaInit() {
#ifdef __i386__
/*	if (hasStreamingSIMDExtensions) {
		VIFfuncTable[0x4].funcU = UNPACK_V2_32_SSE;
		VIFfuncTable[0x4].funcS = UNPACK_V2_32_SSE;
		VIFfuncTable[0x8].funcU = UNPACK_V3_32_SSE;
		VIFfuncTable[0x8].funcS = UNPACK_V3_32_SSE;
		VIFfuncTable[0xC].funcU = UNPACK_V4_32_SSE;
		VIFfuncTable[0xC].funcS = UNPACK_V4_32_SSE;
	}*/
#endif
}

__inline static int _limit( int a, int max ) {
   return ( a > max ? max : a );
}

static void VIFunpack(u32 *data, vifCode *v, int size, const unsigned int VIFdmanum) {
	u32 *dest;
	unsigned int unpackType;
	UNPACKFUNCTYPE func;
	UNPACKPARTFUNCTYPE funcP;
	VIFUnpackFuncTable *ft;
	int *vifUsn;
	VIFregisters *vifRegs;
	VURegs * VU;
	u8 *cdata = (u8*)data;
	int bytes;
	int memsize;

	if (VIFdmanum == 0) {
		VU = &VU0;
		vifUsn = &vif0.usn;
		vifRegs = vif0Regs;
		memsize = 0x1000;
	} else {
		VU = &VU1;
		vifUsn = &vif1.usn;
		vifRegs = vif1Regs;
		memsize = 0x4000;
	}

	v->addr &= memsize-1;
	dest = (u32*)(VU->Mem + v->addr);
  
#ifdef VIF_LOG
   VIF_LOG( "VIF%d UNPACK: Mode=%x, v->size=%d, size=%d, v->addr=%x\n", 
            VIFdmanum, v->cmd & 0xf, v->size, size, v->addr );
#endif

/*	if (vifRegs->cycle.cl > vifRegs->cycle.wl) {
	   SysPrintf( "VIF%d UNPACK: Mode=%x, v->size=%d, size=%d, v->addr=%x\n", 
            VIFdmanum, v->cmd & 0xf, v->size, size, v->addr );
	}*/
	if (v->size != size) {
		SysPrintf("*PCSX2*: warning v->size != size\n");
	}
	if ((v->addr+size*4) > memsize) {
		SysPrintf("*PCSX2*: fixme unpack overflow\n");
	   SysPrintf( "VIF%d UNPACK: Mode=%x, v->size=%d, size=%d, v->addr=%x\n", 
            VIFdmanum, v->cmd & 0xf, v->size, size, v->addr );
	}
//	   SysPrintf( "VIF%d UNPACK: Mode=%x, v->size=%d, size=%d, v->addr=%x\n", 
  //          VIFdmanum, v->cmd & 0xf, v->size, size, v->addr );

	if (size == 0) {
		return;
	}

	_vifRegs = (VIFregisters*)vifRegs;

	// The unpack type
	unpackType = v->cmd & 0xf;
	ft = &VIFfuncTable[ unpackType ];

	// Assigning the normal upack function, the part type is assigned later
	func = *vifUsn ? ft->funcU : ft->funcS;
	funcP = *vifUsn ? ft->funcUpart : ft->funcSpart;

	if (func == NULL) {
		SysPrintf( "UNPACK %x, %x, UNHANDLED %x\n", v->addr, v->size, unpackType );
		return;
	}

	{
		int i;
		for (i=0; i<size; i++) {
			VIF_LOG("[%d] %8.8x\n", i, data[i]);
		}
	}
	// Unpacking
	wl = 0; cl = 0;
	size*= 4;
	if (vifRegs->cycle.cl >= vifRegs->cycle.wl) { /* skipping write */
		while (size >= ((16/ft->gsize)/ft->dsize)) {
//			VIF_LOG("cl %d wl %d; dest=%p cdata=%p\n", cl, wl, dest, cdata);
			if (cl < vifRegs->cycle.wl) { /* unpack one qword */
				bytes = funcP( dest, (u32*)cdata, 4 );
				cdata += bytes;
				size -= bytes;
				wl++;
				if (wl == vifRegs->cycle.wl) {
	   	      		wl = 0;
				}
			}
			dest += 4;
			cl++;
			if (cl == vifRegs->cycle.cl) {
            	cl = 0;
			}
		}
		while (size >= ft->dsize) {
			if (cl < vifRegs->cycle.wl) { /* unpack one qword */
				funcP(dest, (u32*)cdata, (size/ft->dsize)*ft->gsize);
				break;
			}
			dest += 4;
			cl++;
			if (cl == vifRegs->cycle.cl) {
            	cl = 0;
			}
		}
	} else
	if (vifRegs->cycle.cl < vifRegs->cycle.wl) { /* filling write */
		SysPrintf("*PCSX2*: filling write\n");
		while (size >= ft->bsize) {
			if (wl == vifRegs->cycle.wl) {
         		wl = 0;
			}
			func(dest, (u32*)cdata);
			if (wl < vifRegs->cycle.cl) { /* unpack one qword */
				cdata += ft->bsize;
				size -= ft->bsize;
				cl++;
				if (cl == vifRegs->cycle.cl) {
	   	      		cl = 0;
				}
			}
			dest += 12;
			wl++;
		}
	} else {
		while ( size >= ft->bsize ) {
			func(dest, (u32*)cdata);
			dest += 12;
			cdata += ft->bsize;
			size -= ft->bsize;
			cl++; wl++;
			if (cl == vifRegs->cycle.cl) {
				cl = 0;
			}
			if (wl == vifRegs->cycle.wl) {
				wl = 0;
			}
		}
		if (size >= ft->dsize) {
			funcP(dest, (u32*)cdata, (size/ft->dsize)*4);
		}
	}
#ifdef __i386__
	//_emms();
#endif
}

static void vuExecMicro( u32 addr, const unsigned int VIFdmanum ) {
	VURegs * VU;
	void (*_vuExecMicro)();

	if (VIFdmanum == 0) {
		_vuExecMicro = Cpu->ExecuteVU0Block;
		VU = &VU0;
		vif0FLUSH();
	} else {
		_vuExecMicro = Cpu->ExecuteVU1Block;
		VU = &VU1;
		vif1FLUSH();
	}

	VU->vifRegs->itop = VU->vifRegs->itops;

	if (VIFdmanum == 1) {
		VU->vifRegs->top = VU->vifRegs->tops;

		if (VU->vifRegs->stat & 0x80) {
			VU->vifRegs->tops = VU->vifRegs->base + VU->vifRegs->ofst;
			VU->vifRegs->stat &= ~0x80;
		} else {
			VU->vifRegs->tops = VU->vifRegs->base;
			VU->vifRegs->stat |= 0x80;
		}
	}

	if (VIFdmanum == 0) {
		vu0ExecMicro(addr);
	} else {
		vu1ExecMicro(addr);
	}
}

void vif0Init() {
	vif0ch = (DMACh*)&psH[0x8000];
}

int vif0Interrupt() {
	vif0ch->chcr &=~ 0x100;
	hwDmacIrq(DMAC_VIF0);
	vif0Regs->stat&= ~0x1F000000; // FQC=0

	return 1;
}

void vif0FLUSH() {
	while (VU0.VI[REG_VPU_STAT].UL & 0x1) {
		Cpu->ExecuteVU0Block();
	}
}

int VIF0transfer( u32 *data, int size )
{
   int vifNum;

#ifdef VIF_LOG 
   VIF_LOG( "VIF0transfer: size %x (vif0.cmd %x)\n", size, vif0.cmd );
#endif

   while ( size > 0 )
   {
      if ( vif0.cmd )
      {
#ifdef VIF_LOG 
         VIF_LOG("VIFtransferData: cmd %x, size %x, vif0.tag.size %x\n", vif0.cmd, size, vif0.tag.size);
#endif
         if ( ( vif0.cmd & 0x60 ) == 0x60 )
         { // UNPACK
#ifdef VIF_LOG 
            VIF_LOG("UNPACKData: cmd %x, size %x, vif0.tag.size %x\n", vif0.cmd, size, vif0.tag.size);
#endif
            if ( size < vif0.tag.size )
            {
               // CHANGE
               VIFunpack( data, &vif0.tag, size, VIF0dmanum );
               vif0.tag.addr += size * 4;
               vif0.tag.size -= size; 
               size = 0;
            } 
            else 
            {
               // CHANGE
               VIFunpack( data, &vif0.tag, vif0.tag.size, VIF0dmanum );
               data += vif0.tag.size; 
               size -= vif0.tag.size;
               vif0.tag.size = 0;
            }
         } 
         else
         {
            switch (vif0.cmd)
            {
               case 0x20: // STMASK
                  vif0Regs->mask = data[0];
//                if (vif0Regs->mask) SysPrintf("vif0Regs->mask == %x\n", vif0Regs->mask);
                  data++; size--;
                  vif0.tag.size = 0;
                  break;

               case 0x30: // STROW
                  switch (vif0.tag.addr) 
                  {
                     case 0: vif0Regs->r0 = data[0]; break;
                     case 1: vif0Regs->r1 = data[0]; break;
                     case 2: vif0Regs->r2 = data[0]; break;
                     case 3: vif0Regs->r3 = data[0]; break;
                  }
                  data++; size--;
                  vif0.tag.addr++;
                  vif0.tag.size--;
                  break;

               case 0x31: // STCOL
                  switch (vif0.tag.addr)
                  {
                     case 0: vif0Regs->c0 = data[0]; break;
                     case 1: vif0Regs->c1 = data[0]; break;
                     case 2: vif0Regs->c2 = data[0]; break;
                     case 3: vif0Regs->c3 = data[0]; break;
                  }
                  data++; size--;
                  vif0.tag.addr++;
                  vif0.tag.size--;
                  break;

               case 0x4A: // MPG
                  if ( size < vif0.tag.size )
                  {
                     memcpy( VU0.Micro + vif0.tag.addr, data, size * 4 );
					 Cpu->ClearVU0(vif0.tag.addr, size);
                     vif0.tag.addr += size * 4;
                     vif0.tag.size -= size; 
                     size = 0;
                  } 
                  else 
                  {
                     memcpy( VU0.Micro + vif0.tag.addr, data, vif0.tag.size * 4 );
					 Cpu->ClearVU0(vif0.tag.addr, vif0.tag.size);
                     data+= vif0.tag.size; 
                     size-= vif0.tag.size;
                     vif0.tag.size = 0;
                  }
                  break;
            }
         }
         if ( vif0.tag.size <= 0 )
         {
            vif0.cmd = 0;
         }
         continue;
      }

      vif0.cmd = (data[0] >> 24);
		vif0Regs->code = data[0];

#ifdef VIF_LOG 
      VIF_LOG( "VIFtransfer: cmd %x, num %x, imm %x, size %x\n", vif0.cmd, (data[0] >> 16) & 0xff, data[0] & 0xffff, size );
#endif

      if ( vif0.cmd & 0x80 )
      {
#ifdef VIF_LOG
         VIF_LOG( "Interrupt on VIFcmd: %x (INTC_MASK = %x)\n", vif0.cmd, psHu32(INTC_MASK) );
#endif
         hwIntcIrq(VIF0intc);
         return -1;
      }

      if ( ( vif0.cmd & 0x60 ) == 0x60 )
      {
         int vifNum;
         int vl, vn;
         int len;

         vif0FLUSH();

         vl = (vif0.cmd     ) & 0x3;
         vn = (vif0.cmd >> 2) & 0x3;
         vif0.tag.addr = (data[0] & 0x3ff);
         vif0.usn = (data[0] >> 14) & 0x1;
         vifNum = (data[0] >> 16) & 0xff;
         if ( vifNum == 0 )
         { 
            vifNum = 256;
         }

         if ( vif0Regs->cycle.wl <= vif0Regs->cycle.cl )
         {
            len = ((int)ceil( (double)((( 32 >> vl ) * ( vn + 1 )) * vifNum ) / 32 ));
         } 
         else 
         {
            int n = vif0Regs->cycle.cl * (vifNum / vif0Regs->cycle.wl) + 
                    _limit( vifNum % vif0Regs->cycle.wl, vif0Regs->cycle.cl );
            len = ((int)ceil( (double)(( ( 32 >> vl ) * ( vn + 1 )) * n ) / 32 ));
         }
   
         vif0.tag.addr *= 16;
         vif0.tag.cmd  = vif0.cmd;
         vif0.tag.size = len;
         vif0Regs->offset = 0;
      } 
      else
      {
         switch ( vif0.cmd & 0x7F ) 
         {
            case 0x00: // NOP
               vif0.cmd = 0;
               break;

            case 0x01: // STCYCL
               vif0Regs->cycle.cl =  data[0] & 0xff;
               vif0Regs->cycle.wl = (data[0] >> 8) & 0xff;
               vif0.cmd = 0;
               break;

            case 0x04: // ITOP
               vif0Regs->itops = data[0] & 0x3ff;
               vif0.cmd = 0;
               break;

            case 0x05: // STMOD
               vif0Regs->mode = data[0] & 0x3;
               vif0.cmd = 0;
               break;

            case 0x07: // MARK
               vif0Regs->mark = data[0] & 0xffff;
               vif0.cmd = 0;
               break;

            case 0x10: // FLUSHE
               vif0FLUSH();
               vif0.cmd = 0;
               break;

            case 0x14: // MSCAL
            case 0x15: // MSCALF
               vuExecMicro( ( data[0] & 0xffff ) * 8, VIF0dmanum );
               vif0.cmd = 0;
               break;

            case 0x17: // MSCNT
               vuExecMicro( -1, VIF0dmanum );
               vif0.cmd = 0;
               break;

            case 0x20: // STMASK
               break;

            case 0x30: // STROW
               vif0.tag.addr = 0;
               vif0.tag.size = 4;
               break;

            case 0x31: // STCOL
               vif0.tag.addr = 0;
               vif0.tag.size = 4;
               break;

            case 0x4A: // MPG
               vif0FLUSH();
               vifNum = (data[0] >> 16) & 0xff;
               if ( !vifNum )
               {
                  vifNum = 256;
               }
               vif0.tag.addr = ( data[0] & 0xffff ) * 8;
               vif0.tag.size = vifNum * 2;
               break;

            default:
               SysPrintf( "UNKNOWN VifCmd: %x\n", vif0.cmd );
               vif0.cmd = 0;
               if ( ( vif0Regs->err & 0x4 ) == 0 ) 
               {
                  vif0Regs->stat |= 1 << 13;
               }
               break;
         }
      }   
      data++; 
      size--;
   }

   return 0;
}

/* Changed from define to inline. asadr */
void ChainVIF0( DMACh *dma )
{
   if ( dma->qwc )
   {
      u32 * pMem;

      _dmaGetAddr( pMem, dma->madr, VIF0dmanum );
      // CHANGE
      if ( VIF0transfer( pMem, dma->qwc * 4 ) == -1 )
      {
         dma->chcr &=~ ( 1 << 8 );
         hwDmacIrq(VIF0dmanum);
         return;
      } 
   }
}

void dmaVIF0(DMACh *dma) {
   u32 *ptag;
   u32 temp;
   int id;
   BOOL dmaComplete = FALSE;

#ifdef VIF_LOG
   VIF_LOG( "dmaVIF%s chcr = %lx, madr = %lx, qwc  = %lx\n"
            "        tadr = %lx, asr0 = %lx, asr1 = %lx\n",
#ifdef VIF0
               "0",
#else
               "1",
#endif
               dma->chcr, dma->madr, dma->qwc,
               dma->tadr, dma->asr0, dma->asr1 );
#endif

	if (!( dma->chcr & 0x4)) { // Normal Mode 
		ChainVIF0( dma );
		dma->chcr &=~ ( 1 << 8 );
		hwDmacIrq(VIF0dmanum);
		return;
	}

   ChainVIF0( dma );

   // Chain Mode
   while ( !dmaComplete ) // Loop while Dn_CHCR.STR is 1
   {  
      _dmaGetAddr( ptag, dma->tadr, VIF0dmanum );

      id        = (ptag[0] >> 28) & 0x7;
      dma->qwc  = (u16)ptag[0];
      dma->madr = ptag[1];

#ifdef VIF_LOG
      VIF_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, madr=%lx, tadr=%lx\n",
            ptag[1], ptag[0], dma->qwc, id, dma->madr, dma->tadr);
#endif

      // Transfer dma tag if tte is set
      if ( dma->chcr & 0x40 )
      {
         // CHANGE
         if ( VIF0transfer( ptag + 2, 2 ) == -1 )
         {
            break;
         }
      }

      dma->chcr = ( dma->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );

      switch ( id ) 
      {
         case 0: // refe
            ChainVIF0( dma );
            dma->qwc = 0;
            dma->tadr+= 16;
            dmaComplete = TRUE;
            break;

         case 1: // cnt
            dma->madr = dma->tadr + 16;
            ChainVIF0( dma );
            // Set the taddr to the next tag
            dma->tadr += 16 + (dma->qwc * 16);
            break;

         case 2: // next
            temp = dma->madr;
            dma->madr = dma->tadr + 16;
            ChainVIF0( dma );
            dma->tadr = temp;
            break;

         case 3: // ref
         case 4: // refs
            ChainVIF0( dma );
            dma->tadr += 16;
            break;

         case 5: // call
            temp = dma->madr;
            dma->madr = dma->tadr + 16;
            ChainVIF0( dma );

            if ( dma->asr0 == 0 )
            {
               dma->asr0 = dma->tadr + 16 + (dma->qwc * 16);
            }
            else
            {
               dma->asr1 = dma->tadr + 16 + (dma->qwc * 16);
            }
            
            dma->tadr = temp;
            break;

         case 6: // ret
            dma->madr = dma->tadr + 16;
            ChainVIF0( dma );
            
            if ( dma->asr1 != 0 )
            {
               dma->tadr = dma->asr1;
               dma->asr1 = 0;
            } 
            else 
            {
               if ( dma->asr0 != 0 )
               {
                  dma->tadr = dma->asr0;
                  dma->asr0 = 0;
               }
               else
               {
                  dmaComplete = TRUE;
               }
            }
            break;

         case 7: // end
            dma->madr = dma->tadr + 16;
            ChainVIF0( dma );
            dma->qwc = 0;
            dma->tadr = 0;
            dmaComplete = TRUE;
            break;
      }

      if ( ( dma->chcr & 0x80 ) && ( ptag[0] >> 31 ) )
      {
         _dmaGetAddr( ptag, dma->tadr, VIF0dmanum );
         dma->chcr = ( dma->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );

#ifdef VIF_LOG
         VIF_LOG( "dmaIrq Set\n" );
#endif
         break;
      }

   }

   dma->chcr &=~ ( 1 << 8 );
   hwDmacIrq(VIF0dmanum);
}

void vif0Write32(u32 mem, u32 value) {
	if (mem == 0x10003830) { // MARK
		vif0Regs->stat&= ~0x40;
		vif0Regs->mark = value;
	} else
	if (mem == 0x10003810) { // FBRST
//		SysPrintf("VIF0_FBRST write32 %x\n", value);
		if (value & 0x8) {
			dmaVIF0((DMACh*)&psH[0x8000]);
		}
	} else
	if (mem == 0x10003820) { // ERR
		vif0Regs->err = value;
	}
}

void vif0Reset() {
	memset(&vif0, 0, sizeof(vif0));
}

int  vif0Freeze(gzFile f, int Mode) {
	gzfreeze(&vif0, sizeof(vif0));

	return 0;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

int vif1irq=0;

void vif1Init() {
	vif1ch = (DMACh*)&psH[0x9000];
}

int  vif1Interrupt() {
#ifdef VIF_LOG 
	VIF_LOG("vif1Interrupt: %8.8x\n", cpuRegs.cycle);
#endif
	vif1ch->chcr &=~ 0x100;
	hwDmacIrq(DMAC_VIF1);
	vif1Regs->stat&= ~0x1F000000; // FQC=0
	if (vif1irq > 0) {
		vif1irq--;
		hwIntcIrq(VIF1intc);
	}

	return 1;
}

void vif1FLUSH() {
	while (VU0.VI[REG_VPU_STAT].UL & 0x100) {
		Cpu->ExecuteVU1Block();
	}
}

void vif1UNPACK(u32 *data) {
	int vifNum;
    int vl, vn;
    int len;

	vif1FLUSH();

    vl = (vif1.cmd     ) & 0x3;
    vn = (vif1.cmd >> 2) & 0x3;
    vif1.tag.addr = (data[0] & 0x3ff);
    vif1.usn = (data[0] >> 14) & 0x1;
    vifNum = (data[0] >> 16) & 0xff;
    if ( vifNum == 0 ) vifNum = 256;

    if ( vif1Regs->cycle.wl <= vif1Regs->cycle.cl ) {
        len = ((int)ceil( (double)((( 32 >> vl ) * ( vn + 1 )) * vifNum ) / 32 ));
    } else {
        int n = vif1Regs->cycle.cl * (vifNum / vif1Regs->cycle.wl) + 
                _limit( vifNum % vif1Regs->cycle.wl, vif1Regs->cycle.cl );
        len = ((int)ceil( (double)(( ( 32 >> vl ) * ( vn + 1 )) * n ) / 32 ));
    }
   
    if ( ( data[0] >> 15) & 0x1 ) {
        vif1.tag.addr += vif1Regs->tops;
    }
    vif1.tag.addr *= 16;
    vif1.tag.cmd  = vif1.cmd;
    vif1.tag.size = len;
    vif1Regs->offset = 0;
}

void _vif1mpgTransfer(u32 addr, u32 *data, int size) {
/*	SysPrintf("_vif1mpgTransfer addr=%x; size=%x\n", addr, size);
	{
		FILE *f = fopen("vu1.raw", "wb");
		fwrite(data, 1, size*4, f);
		fclose(f);
	}*/
	if (memcmp(VU1.Micro + addr, data, size * 4)) {
		memcpy(VU1.Micro + addr, data, size * 4);
		Cpu->ClearVU1(addr, size);
	}
}

int vif1transferData(u32 *data, int size) {
	int ret=0;

#ifdef VIF_LOG 
	VIF_LOG("VIFtransferData: cmd %x, size %x, vif1.tag.size %x\n", vif1.cmd, size, vif1.tag.size);
#endif
	if ((vif1.cmd & 0x60) == 0x60) { // UNPACK
#ifdef VIF_LOG 
		VIF_LOG("UNPACKData: cmd %x, size %x, vif1.tag.size %x\n", vif1.cmd, size, vif1.tag.size);
#endif
		if (size < vif1.tag.size) {
			VIFunpack(data, &vif1.tag, size, VIF1dmanum);
			vif1.tag.addr += size * 4;
			vif1.tag.size -= size; 
			ret = size;
		} else {
			VIFunpack(data, &vif1.tag, vif1.tag.size, VIF1dmanum);
			ret = vif1.tag.size;
			vif1.tag.size = 0;
		}
	} else {
		switch (vif1.cmd) {
			case 0x20: // STMASK
                vif1Regs->mask = data[0];
//              if (vif1Regs->mask) SysPrintf("vif1Regs->mask == %x\n", vif1Regs->mask);
				ret = 1;
                vif1.tag.size = 0;
                break;

            case 0x30: // STROW
                switch (vif1.tag.addr) {
                    case 0: vif1Regs->r0 = data[0]; break;
                    case 1: vif1Regs->r1 = data[0]; break;
                    case 2: vif1Regs->r2 = data[0]; break;
                    case 3: vif1Regs->r3 = data[0]; break;
                }
				ret = 1;
                vif1.tag.addr++;
                vif1.tag.size--;
                break;

            case 0x31: // STCOL
                switch (vif1.tag.addr) {
                    case 0: vif1Regs->c0 = data[0]; break;
                    case 1: vif1Regs->c1 = data[0]; break;
                    case 2: vif1Regs->c2 = data[0]; break;
                    case 3: vif1Regs->c3 = data[0]; break;
                }
				ret = 1;
                vif1.tag.addr++;
                vif1.tag.size--;
                break;

            case 0x4A: // MPG
                if (size < vif1.tag.size) {
					_vif1mpgTransfer(vif1.tag.addr, data, size);
                    vif1.tag.addr += size * 4;
                    vif1.tag.size -= size; 
                    ret = size;
                } else {
					_vif1mpgTransfer(vif1.tag.addr, data, vif1.tag.size);
					ret = vif1.tag.size;
                    vif1.tag.size = 0;
                }
                break;

            case 0x50: // DIRECT
            case 0x51: // DIRECTHL
                if (size < vif1.tag.size) {
                    GSgifTransfer2(data, size/4);
                    vif1.tag.size-= size;
					ret = size;
                } else {
                    GSgifTransfer2(data, vif1.tag.size/4);
					ret = vif1.tag.size;
                    vif1.tag.size = 0;
                }
				break;
		}
    }
	if (vif1.tag.size <= 0) {
		vif1.cmd = 0;
	}

	return ret;
}

void vif1CMD(u32 *data, int size) {
	int vifNum;
	int vifImm;

	switch ( vif1.cmd & 0x7F ) {
		case 0x00: // NOP
        	vif1.cmd = 0;
            break;

        case 0x01: // STCYCL
            vif1Regs->cycle.cl =  data[0] & 0xff;
            vif1Regs->cycle.wl = (data[0] >> 8) & 0xff;
            vif1.cmd = 0;
            break;

        case 0x02: // OFFSET
            vif1Regs->ofst  = data[0] & 0x3ff;
            vif1Regs->stat &= ~0x80;
            vif1Regs->tops  = vif1Regs->base;
            vif1.cmd = 0;
            break;

        case 0x03: // BASE
            vif1Regs->base = data[0] & 0x3ff;
            vif1.cmd = 0;
            break;

        case 0x04: // ITOP
            vif1Regs->itops = data[0] & 0x3ff;
            vif1.cmd = 0;
            break;

        case 0x05: // STMOD
            vif1Regs->mode = data[0] & 0x3;
            vif1.cmd = 0;
            break;

        case 0x06: // MSKPATH3
            vif1Regs->mskpath3 = (data[0] >> 15) & 0x1; 
            if ( vif1Regs->mskpath3 ) {
                psHu32(GIF_STAT) |= 0x2;
            } else {
                psHu32(GIF_STAT) &= ~0x2;
            }
            vif1.cmd = 0;
            break;

        case 0x07: // MARK
            vif1Regs->mark = data[0] & 0xffff;
            vif1.cmd = 0;
            break;

        case 0x10: // FLUSHE
        case 0x11: // FLUSH
        case 0x13: // FLUSHA
			vif1FLUSH();
            vif1.cmd = 0;
            break;

        case 0x14: // MSCAL
        case 0x15: // MSCALF
            vuExecMicro( ( data[0] & 0xffff ) * 8, VIF1dmanum );
            vif1.cmd = 0;
            break;

        case 0x17: // MSCNT
            vuExecMicro( -1, VIF1dmanum );
            vif1.cmd = 0;
            break;

        case 0x20: // STMASK
            break;

        case 0x30: // STROW
            vif1.tag.addr = 0;
            vif1.tag.size = 4;
            break;

        case 0x31: // STCOL
            vif1.tag.addr = 0;
            vif1.tag.size = 4;
            break;

        case 0x4A: // MPG
			vif1FLUSH();
            vifNum = (data[0] >> 16) & 0xff;
            if (vifNum == 0) vifNum = 256;
            vif1.tag.addr = ( data[0] & 0xffff ) * 8;
            vif1.tag.size = vifNum * 2;
            break;

        case 0x50: // DIRECT
        case 0x51: // DIRECTHL
			if ((size & 1) == 0) {
			    vif1.cmd = 0; break;
			}
            vifImm = data[0] & 0xffff;
            if (vifImm == 0) {
				vif1.tag.size = 65536;
			} else {
				vif1.tag.size = vifImm * 4;
			}
            break;

        default:
			SysPrintf( "UNKNOWN VifCmd: %x\n", vif1.cmd );
            vif1.cmd = 0;
            if ((vif1Regs->err & 0x4) == 0) {
            	vif1Regs->stat |= 1 << 13;
            }
            break;
	}
}


int VIF1transfer(u32 *data, int size) {
	int ret;

#ifdef VIF_LOG 
	VIF_LOG( "VIF1transfer: size %x (vif1.cmd %x)\n", size, vif1.cmd );
#endif

/*	{
		int i;
		for (i=0; i<size; i++) {
#ifdef VIF_LOG 
		   VIF_LOG( "[%d]: %x\n", i, data[i] );
#endif
		}
	}*/

	vif1irq = 0;
	while (size > 0) {
		if (vif1.cmd) {
			ret = vif1transferData(data, size);
			data+= ret; size-= ret;
			continue;
		}

/*		if (irq) {
//			Log=1;
			hwIntcIrq(VIF1intc);
			return tsize - size;
		}*/

		vif1.cmd = (data[0] >> 24);
		vif1Regs->code = data[0];

#ifdef VIF_LOG 
		VIF_LOG( "VIFtransfer: cmd %x, num %x, imm %x, size %x\n", vif1.cmd, (data[0] >> 16) & 0xff, data[0] & 0xffff, size );
#endif

		if (vif1.cmd & 0x80) {
#ifdef VIF_LOG
			VIF_LOG( "Interrupt on VIFcmd: %x (INTC_MASK = %x)\n", vif1.cmd, psHu32(INTC_MASK) );
#endif
			vif1irq++;
		}

		if ((vif1.cmd & 0x60) == 0x60) {
			vif1UNPACK(data);
		} else {
			vif1CMD(data, size);
		}
		data++; 
		size--;
	}

	if (vif1irq > 0) {
		return -2;
	}
	return 0;
}

int  _VIF1chain() {
	u32 *pMem;
	u32 qwc = vif1ch->qwc;
	u32 ret;

	if (vif1ch->qwc == 0) return 0;

	pMem = dmaGetAddr(vif1ch->madr);
	if (pMem == NULL) return -1;

	ret = VIF1transfer(pMem, vif1ch->qwc*4);
	vif1ch->madr+= qwc*16;
	vif1ch->qwc-= qwc;
	cycles+= qwc / BIAS; /* guessing */
	return ret;
}

void dmaVIF1(DMACh *dma) {
	u32 *ptag;
	u32 temp,finalAddress;
	int id;
	int done=0;

#ifdef VIF_LOG
	VIF_LOG("dmaVIF1 chcr = %lx, madr = %lx, qwc  = %lx\n"
			"        tadr = %lx, asr0 = %lx, asr1 = %lx\n",
			dma->chcr, dma->madr, dma->qwc,
			dma->tadr, dma->asr0, dma->asr1 );
#endif
	if (vif1irq > 0) {
		vif1irq--;
		hwIntcIrq(VIF1intc);
		return;
	}

	if (((psHu32(DMAC_CTRL) & 0xC) == 0x8) &&
        (dma->tadr == psHu32(0xe050))) { // VIF MFIFO
		dma->chcr &=~ ( 1 << 8 );
		hwDmacIrq(VIF1dmanum);
		return;
	}

	cycles = 0;
	vif1Regs->stat|= 0x10000000; // FQC=16

	if (!(dma->chcr & 0x4)) { // Normal Mode 
		if ((dma->chcr & 0x1)) { // to Memory
			_VIF1chain();
		} else { // from Memory
			u64 *pMem;
			int size;

			_dmaGetAddr( (u32*)pMem, dma->madr, VIF1dmanum );
			for (size=dma->qwc; size>0; size--) {
				if (size > 1) GSreadFIFO((u64*)&psH[0x5000]);
				pMem[0] = psHu64(0x5000);
				pMem[1] = psHu64(0x5008); pMem+= 2;
			}
			vif1Regs->stat&= ~0x1f000000;
/*         GSgifTransfer2( pMem, dma->qwc );
         if ( ( vif1Regs->stat & 0x800000 ) && ( dma->qwc < 16 ) )
         {
            vif1Regs->addr = 0;
            vif1Regs->stat+= 0x1000000;
            // MFIFO seems to go though FIFO
			// so store the transfer at VIF1 FIFO
			// which can be reading the VIF1 FIFO later
            memcpy(&psH[0x5000], pMem, dma->qwc*16);
         }*/
		}

		INT(1, cycles);
		return;
	}

	if (_VIF1chain() != 0) {
		INT(1, cycles);
		return;
	}

	// Chain Mode
	while (done == 0) {
		ptag = (u32*)dmaGetAddr(dma->tadr);
		if (ptag == NULL) {
			psHu32(DMAC_STAT)|= 1<<15;
			break;
		}

		id        = (ptag[0] >> 28) & 0x7;
		dma->qwc  = (u16)ptag[0];
		dma->madr = ptag[1];

#ifdef VIF_LOG
		VIF_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, madr=%lx, tadr=%lx\n",
				ptag[1], ptag[0], dma->qwc, id, dma->madr, dma->tadr);
#endif

      // Transfer dma tag if tte is set
		if (dma->chcr & 0x40) {
			if (VIF1transfer(ptag + 2, 2) == -1) break;
			cycles++;
		}

		dma->chcr = ( dma->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );

		switch (id) {
			case 0: // refe
				dma->tadr += 16;
				done = 1;
				break;

			case 1: // cnt
				dma->madr = dma->tadr + 16;
				// Set the taddr to the next tag
				dma->tadr += 16 + (dma->qwc * 16);
				break;

			case 2: // next
				temp = dma->madr;
				dma->madr = dma->tadr + 16;
				dma->tadr = temp;
				break;

			case 3: // ref
			case 4: // refs
				dma->tadr += 16;
				break;

			case 5: // call
				temp = dma->madr;
				finalAddress = dma->tadr + 16 + (dma->qwc * 16);

				if (dma->asr0 == 0) 
					dma->asr0 = finalAddress;
				else
					dma->asr1 = finalAddress;

				dma->madr = dma->tadr + 16;
				dma->tadr = temp;
				break;

			case 6: // ret
				dma->madr = dma->tadr + 16;
				if (dma->asr1 != 0) {
					dma->tadr = dma->asr1;
					dma->asr1 = 0;
				} else {
					if(dma->asr0 != 0) {
						dma->tadr = dma->asr0;
						dma->asr0 = 0;
					} else {
						dma->tadr = 0;
						done = 1;
					}
				}
				break;

			case 7: // end
				dma->madr = dma->tadr + 16;
				dma->tadr+= 16;
				done = 1;
				break;
		}
		if (_VIF1chain() != 0) break;

/*		if ( ( dma->chcr & 0x80 ) && ( ptag[0] >> 31 ) ) {
			_dmaGetAddr( ptag, dma->tadr, VIF1dmanum );
//			dma->chcr = ( dma->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );

#ifdef VIF_LOG
			VIF_LOG( "dmaIrq Set\n" );
#endif
			break;
		}*/
	}

	INT(1, cycles);
}

void vif1Write32(u32 mem, u32 value) {
	if (mem == 0x10003c30) { // MARK
		vif1Regs->stat&= ~0x40;
		vif1Regs->mark = value;
	} else
	if (mem == 0x10003c10) { // FBRST
//		SysPrintf("VIF1_FBRST write32 %x\n", value);
		if (value & 0x8) {
			dmaVIF1((DMACh*)&psH[0x9000]);
		}
/*		if (value & 0x4) {
			dmaVIF1((DMACh*)&psH[0x9000]);
		}*/
	} else
	if (mem == 0x10003c20) { // ERR
		vif1Regs->err = value;
	}
}

void vif1Reset() {
	memset(&vif1, 0, sizeof(vif1));
}

int  vif1Freeze(gzFile f, int Mode) {
	gzfreeze(&vif1, sizeof(vif1));

	return 0;
}




